/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005-2010, Red Hat Middleware LLC, and individual contributors
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */


package org.jboss.messaging.core.impl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;

import org.jboss.logging.Logger;
import org.jboss.messaging.core.contract.Message;

/**
 * A DeliveringCounter: keeps track of delivering count and messages.
 * 
 * Because of the asynchronous nature of message delivery and acknowledge, message acknowledgments may
 * come in before corresponding counting increment incurred with message delivery. For example,
 * Message m1 is being delivered to a consumer and then is put to this counter. Normally the ack of m1
 * comes later and causes the counter to decrement. m1 will be correctly added and removed.
 * However, there is a minor chance that the ack of m1 comes in before it is put into this counter.
 * Therefore when decrement() is called m1 is not there. Later m1 is put into the counter and will never
 * be cleared because it has missed the ack. This would be fine with a pure delivering count (an integer)
 * but not when retrieve messages being delivered. In above case, an acked message m1 will be returned 
 * when being asked for even if it has be acked.
 * 
 * Also, adding in-processing message retrieval functionality entails more data structure operations for 
 * message references tracking, thus bringing extra performance cost. This shouldn't be necessary for users
 * who doesn't need this functionality. 
 * 
 * So in the implementation, we use three variables:
 * 
 * - an AtomicInteger to purely track the delivering count;
 * - an ConcurrentHashMap to hold messages being delivered;
 * - an ConcurrentLinkedQueue to hold messages that are being acknowledged but not yet put in the above map.
 *
 * In addition, a int flag to control the functionality level:
 * 
 * - LEVEL_NONE denotes totally disable delivering counter.
 * - LEVEL_COUNTER denotes delivering count only. This is the default level.
 * - LEVEL_ALL denotes full delivering counter functionality.
 * 
 *
 * @author <a href="mailto:hgao@redhat.com">Howard Gao</a>
 * 
 * Created Aug 18, 2010 1:27:46 PM
 */
public class DeliveringCounter
{
   public static final int LEVEL_NONE = 0;
   public static final int LEVEL_COUNTER = 1;
   public static final int LEVEL_ALL = 255;
   
   private static final Logger log = Logger.getLogger(DeliveringCounter.class);
   
   private int level = LEVEL_COUNTER;
   
   private ConcurrentHashMap<Long, Message> msgCount;
   private ConcurrentLinkedQueue<Message> earlyMsgs;
   private AtomicInteger counter;

   public DeliveringCounter()
   {
      this(LEVEL_COUNTER);
   }
   
   public DeliveringCounter(int lv)
   {
      initLevel(lv);
   }
   
   private void initLevel(int lv)
   {
      level = lv;
      log.debug("Delivering Counter enabled: " + level);
      if (level == LEVEL_NONE) return;
      counter = new AtomicInteger(0);
      if (level == LEVEL_ALL)
      {
         msgCount = new ConcurrentHashMap<Long, Message>();
         earlyMsgs = new ConcurrentLinkedQueue<Message>();
      }      
   }
   
   public void decrement(Message acked)
   {
      if (level == LEVEL_NONE) return;
      counter.decrementAndGet();
      if (level != LEVEL_ALL) return;
      Message m = msgCount.remove(acked.getMessageID());
      if (m == null)
      {
         earlyMsgs.add(acked);
      }
   }

   public void increment(Message msg)
   {
      if (level == LEVEL_NONE) return;
      counter.incrementAndGet();
      if (level != LEVEL_ALL) return;
      msgCount.put(msg.getMessageID(), msg);
   }

   public int get()
   {
      if (level == LEVEL_NONE) return 0;
      return counter.get();
   }

   public void reset()
   {
      if (level == LEVEL_NONE) return;
      counter.set(0);
      if (level != LEVEL_ALL) return;
      earlyMsgs.clear();
      msgCount.clear();
   }

   public synchronized List getMessages()
   {
      if (level != LEVEL_ALL) return new ArrayList<Message>();
      Iterator<Message> iterId = earlyMsgs.iterator();
      while (iterId.hasNext())
      {
         Message acked = iterId.next();
         Message m = msgCount.remove(acked.getMessageID());
         if (m != null)
         {
            iterId.remove();
         }
      }
      Iterator<Message> iter = msgCount.values().iterator();
      List<Message> list = new ArrayList<Message>();
      while(iter.hasNext()) {
         list.add(iter.next());
      }
      return list;
   }
   
   //for tests only
   public void setLevel(int lv)
   {
      initLevel(lv);
   }

   //for tests only
   public int getLevel()
   {
      return level;
   }

   public static String toString(int lvl) throws Exception
   {
      if (lvl == LEVEL_NONE)
      {
         return "NONE";
      }
      if (lvl == LEVEL_COUNTER)
      {
         return "COUNTER";
      }
      if (lvl == LEVEL_ALL)
      {
         return "ALL";
      }
      throw new Exception("Invalid value: " + lvl);
   }

   public static int toLevel(String lvl) throws Exception
   {
      if ("NONE".equals(lvl))
      {
         return LEVEL_NONE;
      }
      if ("COUNTER".equals(lvl))
      {
         return LEVEL_COUNTER;
      }
      if ("ALL".equals(lvl))
      {
         return LEVEL_ALL;
      }
      throw new Exception("Invalid value: \"" + lvl + "\"");
   }
}
